#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import os
import logging
import subprocess
import re
import gzip
import shutil

from yaml import load, dump
import magic

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

from coodeer.configuration.coodeer import Settings
from coodeer.configuration.coodeer import LATEST_REPOSITORY_SPECIFICATION_KEY
from coodeer.controller.s3_access import S3ObjectsControl

PATTERN_PACKAGES_FILENAME = r"^Filename\: (?P<path>.*?)$"
REGEX_PACKAGES_FILENAME = re.compile(PATTERN_PACKAGES_FILENAME)


def mk_rsync_call_args(source, destination):
    args = ["rsync", "-avrH", "--delete", source, destination]

    return args


def update_repository_metadata(repo_path, root):
    log = logging.getLogger(__name__)

    scanpackages_args = [
        "dpkg-scanpackages",
        "--multiversion",
        ".",
    ]
    packages_out = subprocess.check_output(scanpackages_args, cwd=repo_path)
    packages_gz = os.path.join(repo_path, "Packages.gz")

    with gzip.open(packages_gz, "wt") as tgt:
        for line in packages_out.decode("utf-8").split("\n"):
            matcher = REGEX_PACKAGES_FILENAME.match(line)

            if matcher:
                f_path = matcher.groupdict()["path"]
                abs_path = os.path.join(repo_path, f_path)
                rel_path = os.path.relpath(abs_path, root)
                line = f"Filename: {rel_path}"
                log.debug(f"Mangled filename {f_path!r} -> {rel_path!r}")

            tgt.write(line)
            tgt.write("\n")


class RepositoryControl:
    def __init__(self, *args, **kwargs):
        self.log = logging.getLogger(__name__)
        self.cfg = Settings()
        self.soc = S3ObjectsControl()

    def upload(self, path, blacklisted=None, whitelisted_content_types=None):
        items = set()

        if os.path.isfile(path):
            items.add((os.path.basename(path), os.path.abspath(path)))
        elif os.path.isdir(path):
            for item in os.listdir(path):
                try:
                    if item in blacklisted:
                        self.log.info(f"BLACKLISTED: {item}")
                        continue
                except TypeError:
                    pass

                abs_path = os.path.abspath(os.path.join(path, item))
                items.add((item, abs_path))

        for item, abs_path in items:
            try:
                mime_type = magic.from_file(abs_path, mime=True)
            except NameError:
                self.log.warning("The magic library appears to be missing?")
                mime_type = "application/octet-stream"

            try:
                if mime_type not in whitelisted_content_types:
                    self.log.info(f"IGNORED: {item}")
                    continue
            except TypeError:
                pass

            rel_path = item
            self.log.info(f"+ {abs_path} {mime_type} --> {rel_path}")

            try:
                self.soc.push(rel_path, abs_path, mime_type)
            except Exception as exc:
                self.log.error(exc)

    def get_latest_repository_specification(self, repositories_root):
        return self.soc.pull(
            LATEST_REPOSITORY_SPECIFICATION_KEY, root=repositories_root
        )

    def validate_repository_specification(self, data):
        errors = list()

        for required_key in ("repositories_mapping", "published"):
            if required_key not in data:
                errors.append(
                    f"Specification is missing required key {required_key!r}!"
                )

        if not errors:
            repositories_mapping = data["repositories_mapping"]
            published = data["published"]

            for package, keys in published.items():
                for key in sorted(keys):
                    if key not in repositories_mapping:
                        errors.append(
                            f"Package {package!r} published to undefined repository {key!r}!"
                        )

        if errors:
            for item in errors:
                self.log.error(item)

            raise ValueError("Invalid Specification!")

        return data

    def put_latest_repository_specification(self, spec_file):
        with open(spec_file, "r") as src:
            spec = load(src, Loader=Loader)

        self.validate_repository_specification(spec)

        self.soc.push(
            LATEST_REPOSITORY_SPECIFICATION_KEY, spec_file, "application/yaml"
        )

    def create_repository_specification(self, data, spec_file):
        self.validate_repository_specification(data)

        with open(spec_file, "w") as tgt:
            tgt.write(dump(data, Dumper=Dumper))

        return spec_file

    def mk_repositories(self, repositories_root, spec_file):
        with open(spec_file, "r") as src:
            spec = load(src, Loader=Loader)

        self.validate_repository_specification(spec)

        pool_root = os.path.join(repositories_root, "pool")
        repositories_mapping = spec["repositories_mapping"]

        self.log.info(
            f"Creating repositories in {repositories_root!r}, pool in {pool_root}"
        )
        self.log.debug(f" Pool: {pool_root!r}")
        os.makedirs(pool_root, exist_ok=True)

        folder_mapping = dict()
        for repo_key, path in sorted(repositories_mapping.items()):
            folder_path = os.path.join(repositories_root, path)
            self.log.info(f" {repo_key}: {folder_path!r}")
            folder_mapping[repo_key] = folder_path

            if os.path.isdir(folder_path):
                self.log.debug(f" {repo_key}: Dropping {folder_path}")
                shutil.rmtree(folder_path)

            os.makedirs(folder_path, exist_ok=True)

        for package, published_to in sorted(spec["published"].items()):
            targets = []

            for repo_key in published_to:
                target_path = os.path.join(folder_mapping[repo_key], package)
                targets.append((repo_key, target_path))

            self.log.info(f"Package {package!r}")

            try:
                pool_target = self.soc.pull(package, pool_root)
            except KeyError as kexc:
                self.log.warning(f" Missing! IGNORED.")
                continue

            self.log.debug(f" Source: {pool_target!r}")

            for repo_key, target in sorted(targets):
                self.log.info(f" {repo_key}: {target}")

                try:
                    os.link(pool_target, target)
                except FileExistsError:
                    os.unlink(target)
                    os.link(pool_target, target)

        for repo_path in folder_mapping.values():
            update_repository_metadata(repo_path, repositories_root)

        self.log.info(
            f"You could call >> {' '.join(mk_rsync_call_args(repositories_root, '$TARGET'))} << for synchronising now."
        )

        return repositories_root



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import sys
import logging

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)-8s %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

LOG = logging.getLogger("coodeer")

import click

from coodeer.configuration.coodeer import CoodeerSettings
from coodeer.controller.repository import RepositoryControl

CONTEXT_SETTINGS = dict(
    help_option_names=["-h", "--help"],
)

pass_client = click.make_pass_decorator(RepositoryControl, ensure=True)

cfg_defaults = CoodeerSettings()


@click.group()
@click.option(
    "--bucket",
    show_default=True,
    default=cfg_defaults.bucket,
    help="Source Bucket",
)
@click.pass_context
def cli(ctx, **kwargs):
    ctx.obj = RepositoryControl(kwargs["bucket"])


@cli.command("upload", context_settings=CONTEXT_SETTINGS)
@pass_client
@click.argument("packages", nargs=-1)
def coodeer_upload(ctx, **kwargs):
    """
    Upload debian packages to source bucket
    """
    whitelisted_content_types = [
        "application/vnd.debian.binary-package",
        "application/pgp-signature",
    ]

    for path in kwargs["packages"]:
        try:
            ctx.upload(
                path, whitelisted_content_types=whitelisted_content_types
            )
        except Exception as exc:
            LOG.error(exc)
            sys.exit(1)

    sys.exit(0)


@cli.command("create", context_settings=CONTEXT_SETTINGS)
@pass_client
@click.option(
    "repository_specification",
    "--repository-specification",
    help="Repository specification file (YAML)",
)
@click.argument("repositories_root")
def coodeer_create(ctx, **kwargs):
    """
    Create local repositories
    """
    repositories_root = kwargs["repositories_root"]

    if not os.path.isdir(repositories_root):
        try:
            os.makedirs(repositories_root)
        except Exception as exc:
            LOG.error(exc)
            sys.exit(4)

    if not kwargs.get("repository_specification"):
        try:
            repository_specification = ctx.get_latest_repository_specification(
                repositories_root
            )
        except KeyError:
            LOG.error("No repository specification available!")
            sys.exit(3)
    else:
        repository_specification = kwargs.get("repository_specification")

    try:
        ctx.mk_repositories(repositories_root, repository_specification)
    except Exception as exc:
        LOG.error(exc)
        sys.exit(1)

    sys.exit(0)


@cli.command("publish", context_settings=CONTEXT_SETTINGS)
@pass_client
@click.argument("repository_specification")
def coodeer_publish(ctx, **kwargs):
    """
    Publish repository specification
    """
    repository_specification = kwargs.get(
        "repository_specification",
        help="Repository specification file (YAML)",
    )

    try:
        ctx.put_latest_repository_specification(repository_specification)
    except ValueError as vexc:
        LOG.error(vexc)
        sys.exit(2)
    except Exception as vexc:
        LOG.error(vexc)
        sys.exit(1)

    sys.exit(0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from pydantic import BaseSettings


LATEST_REPOSITORY_SPECIFICATION_KEY = "00-spec/latest.yaml"


class CoodeerSettings(BaseSettings):
    bucket: str = "coodeer"

    class Config:
        env_prefix = "COODEER_"  # defaults to no prefix, i.e. ""


class AmazonWebServicesSettings(BaseSettings):
    aws_access_key_id: str
    aws_default_region: str
    aws_secret_access_key: str

    class Config:
        fields = {
            "aws_access_key_id": {
                "env": "AWS_ACCESS_KEY_ID",
            },
            "aws_default_region": {
                "env": "AWS_DEFAULT_REGION",
            },
            "aws_secret_access_key": {
                "env": "AWS_SECRET_ACCESS_KEY",
            },
        }


class Settings(CoodeerSettings, AmazonWebServicesSettings, BaseSettings):
    language: str = "C"
    lc_all: str = "C"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

HEAD response

    * see https://docs.aws.amazon.com/AmazonS3/latest/API/RESTCommonResponseHeaders.html

"""
import os
import logging
import hashlib

import boto3
import botocore

logging.getLogger("botocore").setLevel(logging.WARNING)

from coodeer.configuration.coodeer import Settings


def md5_sum(local_fn):
    """
    Return MD5 sum of file *local_fn*.
    """
    with open(local_fn, "rb") as handle:
        md5sum = hashlib.md5(handle.read()).hexdigest()

    return md5sum


class S3ObjectsControl:
    """
    Asset's up-/downloading Controller.
    (Based on https://github.com/doubleO8/sure-shot-static/blob/master/sure_shot_static/intergalactic.py)


    Attributes:
        log (logging.Logger): logger instance
        bucket (str): Bucket name
        base_url (str): Bucket's base URL
        base_arn (str): Bucket's base ARN
    """

    def __init__(self, *args, **kwargs):
        self.log = logging.getLogger(__name__)
        self.cfg = Settings()
        self.bucket = kwargs.get("bucket", self.cfg.bucket)
        self._session = boto3.session.Session()
        self.client = boto3.client("s3")

    @property
    def base_url(self):
        return "https://{bucket}.s3.{region_name}.amazonaws.com".format(
            bucket=self.bucket, region_name=self._session.region_name
        )

    @property
    def base_arn(self):
        return "arn:aws:s3:::{bucket}".format(bucket=self.bucket)

    def head(self, rel_path):
        try:
            return self.client.head_object(Bucket=self.bucket, Key=rel_path)
        except botocore.exceptions.ClientError as bex:
            if bex.response["ResponseMetadata"]["HTTPStatusCode"] == 404:
                raise KeyError(rel_path)
            elif bex.response["ResponseMetadata"]["HTTPStatusCode"] == 403:
                self.log.error(
                    "You Gotta Fight For Your Right To ... {what:8} {rel_path!r}: {Code} {Message}".format(
                        what="get head",
                        rel_path=rel_path,
                        **bex.response["Error"],
                    )
                )
            else:
                raise

        return None

    def pull(self, rel_path, root, **kwargs):
        need_pull = True
        abs_path = os.path.abspath(os.path.join(root, rel_path))

        if os.path.exists(abs_path):
            try:
                head_response = self.head(rel_path)
            except KeyError:
                raise

            if head_response:
                need_pull = False

                try:
                    md5_digest = md5_sum(abs_path)
                    key = "ETag"
                    self.log.debug(
                        "   {:20}: me={!r} -- them={!r}".format(
                            key, md5_digest, head_response.get(key)
                        )
                    )

                    if md5_digest not in head_response["ETag"]:
                        need_pull = True
                except Exception as exc:
                    self.log.warning(
                        "MD5 sum/ETag comparison failed: {!s}".format(exc)
                    )

        if need_pull:
            parent = os.path.dirname(abs_path)
            os.makedirs(parent, exist_ok=True)

            s3_resource = boto3.resource("s3")

            try:
                s3_resource.Object(self.bucket, rel_path).download_file(
                    abs_path
                )
            except botocore.exceptions.ClientError as bex:
                if bex.response["ResponseMetadata"]["HTTPStatusCode"] == 404:
                    raise KeyError(rel_path)
                else:
                    raise

        return abs_path

    def push(self, rel_path, abs_path, mime_type, **kwargs):
        """
        Upload an object to the bucket.

        Args:
            rel_path (str): relative path (target)
            abs_path (str): absolute path (local source)
            mime_type (str): MIME type

        Returns:
            bool: ``True`` if operation succeeded
        """
        need_push = True
        head_response = None

        upload_args = {
            "Key": rel_path,
            "Bucket": self.bucket,
            "ContentType": mime_type,
        }

        try:
            head_response = self.head(rel_path)
        except KeyError:
            pass

        if head_response:
            need_push = False

            try:
                md5_digest = md5_sum(abs_path)
                key = "ETag"
                self.log.debug(
                    "   {:20}: me={!r} -- them={!r}".format(
                        key, md5_digest, head_response.get(key)
                    )
                )

                if md5_digest not in head_response["ETag"]:
                    need_push = True
            except Exception as exc:
                self.log.warning(
                    "MD5 sum/ETag comparison failed: {!s}".format(exc)
                )

            if need_push is False:
                same_size = (
                    os.path.getsize(abs_path) == head_response["ContentLength"]
                )
                if not same_size:
                    self.log.debug("   Size differs ...")
                    need_push = True

        if not need_push:
            self.log.info("   No update needed ...")
            return True

        succeeded = False
        with open(abs_path, "rb") as src:
            try:
                self.client.put_object(Body=src, **upload_args)
                succeeded = True
            except botocore.exceptions.ClientError as bex:
                if bex.response["ResponseMetadata"]["HTTPStatusCode"] == 403:
                    self.log.error(
                        "You Gotta Fight For Your Right To ... {what:8} {rel_path!r}: {Code} {Message}".format(
                            what="put",
                            rel_path=rel_path,
                            **bex.response["Error"],
                        )
                    )
                else:
                    raise

        if succeeded:
            self.log.info(" = {:s}/{:s}".format(self.base_url, rel_path))

        return succeeded
